class StressStrain:
    def __init__(self, raw, sample_area, initial_length):
        self.raw = raw

        strain = self.getStrain(sample_area)
        stress = self.getStress(initial_length)
        
        import pandas as pd
        self.data = pd.DataFrame({'strain': strain, 'stress': stress})
    
    def getStress(self, sample_area):
        return self.raw['N'] / float(sample_area) * 1_000_000 / 1000
    
    def getStrain(self, initial_length):
        from numpy import log
        length = self.raw['mm'] + float(initial_length)
        return log(length / initial_length)
    





class Fit:
    import pandas as pd
    from .analyzers import StressStrain
    
    def __init__(self, data: StressStrain, cut_off: bool = True, fit_window: int = 200):
        
        self.data = data.data
        self.max_stress_index = self.maxStressIndex()
        
        if cut_off:
            self.cutOffData()

        self.fit_results = self.getBestFit(fit_window)
        self.results = self.getResults()
        
        
    def maxStressIndex(self):
        max_stress_row = self.data[self.data['stress'] == self.data['stress'].max()]
        return max_stress_row.index[0]
        
    def getResults(self) -> pd.DataFrame:
        import pandas as pd
        
        maxValues = self.data.iloc[self.max_stress_index]
        yieldValues = self.getYield()
        return pd.DataFrame({
            'Max Stress [kPa]': [maxValues['stress']],
            'Max Strain': [maxValues['strain']], 
            'Young Modulus [kPa]': [self.fit_results.slope],
            'Intercept [kPa]': [self.fit_results.intercept],
            'Yield Stress [kPa]': [yieldValues['stress']],
            'Yield Strain': [yieldValues['strain']],})
    

    def cutOffData(self):
        self.data = self.data.iloc[0:self.max_stress_index + 1]
        return None
    
    def getBestFit(self, window = 200):
        import pandas as pd
        import numpy as np
        
        window = 200
        step = 10

        fit_results = pd.DataFrame()

        sizeData = len(self.data[self.data['strain'] < 0.1])

        if window / sizeData > 0.2:
            window = max([int(sizeData * 0.2), 30])
            step = 5

        left = 0
        right = window

        while True:

            if right > sizeData: 
                break

            data = self.data[left:right]

            slope, intercept = np.polyfit(data['strain'], data['stress'], 1)
            y_exp = slope * data['strain'] + intercept
            error = (data['stress'] - y_exp)**2
            df = pd.DataFrame({'strain':  data[['strain']].iloc[int(window / 2)].values, 'slope': slope, 'intercept': intercept, 'error': error.sum()})

            fit_results = pd.concat([fit_results, df], axis=0)

            left += step
            right += step
        
        bestFit = fit_results[fit_results['slope'] == fit_results['slope'].max()].iloc[0]
        return bestFit
    
    def getYield(self):
        df = self.data.copy()
        df['error'] = abs(df['stress'] - (df['strain']-0.02)*self.fit_results.slope - self.fit_results.intercept)
        yieldValues = df[df['error'] == df['error'].min()].iloc[0]
        return yieldValues


# Copyright 2023 Paride Azzari
#
# Licensed under the MIT License. See LICENSE

def analyzeDirectory(sample_area):
    '''
    Give a folder path as a relative or absolute path, the script will analyze all the .TAR files found in the directory and return a DataFrame containing the results.
    Leaving returns the Current Working Directory.
    '''
    from colorama import init, Fore
    init()

    print(Fore.RED + "---------------------------")
    print(Fore.RED + "--------- WARNING ---------")
    print(Fore.RED + "---------------------------")
    print(Fore.GREEN + "!!   OUTDATED VERSION   !!")
    print(Fore.WHITE + 'Please download the latest version at: ')
    print("https://github.com/azzarip/extrudion")
    print(Fore.RED + "---------------------------")
    print(Fore.RED + "--------- WARNING ---------")
    print(Fore.RED + "---------------------------")
    print(Fore.GREEN + "!!   OUTDATED VERSION   !!")
    print(Fore.WHITE + 'Please download the latest version at: ')
    print("https://github.com/azzarip/extrudion")
    print(Fore.RED + "---------------------------")
    print(Fore.RED + "--------- WARNING ---------")
    print(Fore.RED + "---------------------------")
    print(Fore.GREEN + "!!   OUTDATED VERSION   !!")
    print(Fore.WHITE + 'Please download the latest version at: ')
    print("https://github.com/azzarip/extrudion")
    print(Fore.RED + "---------------------------")
    print(Fore.WHITE)
    return 

def start(folder_path: str = '.'):
    '''
    Give a folder path as a relative or absolute path, the script will analyze all the .TAR files found in the directory and return a DataFrame containing the results.
    Leaving returns the Current Working Directory.
    '''
    while True:
        sample_area = input("Enter the Surface of the Sample [mm2]: ")
    
        if sample_area.replace('.', '', 1).isdigit():
            break
        else:
            print("Invalid input. Please enter a valid number.")
 
    while True:
        initial_length = input("Enter the Initial Length of the Sample [mm]: ")
    
        if initial_length.replace('.', '', 1).isdigit():
            break
        else:
            print("Invalid input. Please enter a valid number.")
 
    from .files import TRAFolder
    TRAFolder(folder_path).analyze({'sample_area': float(sample_area), 'initial_length': float(initial_length)})
        
    


class TRAFolder:
    class FolderNotFound(Exception):
        pass
    
    def __init__(self, folder_path: str):
        self.folder_path = folder_path
        
        self.file_list =self.getTRAFiles()

        import pandas as pd
        self.results = pd.DataFrame()
        
        import os
        if not os.path.exists(folder_path + '/plots'):
            os.makedirs(folder_path + '/plots') 
    
    def analyze(self, options = {}):
        if not self.file_list:
            return

        import pandas as pd
        
        for file in self.file_list:
            result = TRAFile(file, self.folder_path).analyze(options)
            result['File'] = file
            self.results = pd.concat([self.results, result])
            
        self.printCopyrights()
        return self.results.set_index('File')
    
    def getTRAFiles(self):
        import os
            
        try:
            dir_list = os.listdir(self.folder_path)
            
            try:
                file_list = [filename for filename in dir_list if filename.endswith('.TRA')]

                if not file_list:
                    print('No .TRA files found in the folder.')
                    return []
                else:
                    return file_list
            
            except FileNotFoundError:
                print('An error occurred while filtering files.')
                raise
            
        except FileNotFoundError:
            raise TRAFolder.FolderNotFound('Folder not found') 
        
    def printCopyrights(self):
        print("*********************************************")
        print("                EXTRUDION")
        print("*********************************************")
        print()
        print("by Paride Azzari (C) 2024")
        print()
        print("info on: github.com/azzarip/extrudion")
        print("*********************************************")
        print("RESULTS:")
        print()
        print("Analysis.csv contains the analyzed data")
        print()
        print("The plots folder contains all the figures")
        print("*********************************************")
        print()
    

class TRAFile:
    
    def __init__(self, file: str, folder_path: str):
        import os
        import pandas as pd
        self.filename = file.replace('.TRA', '')
        
        if folder_path:
            self.filepath = os.path.join(folder_path, file)
        else:
            self.filepath = file
            
            
        df = pd.read_table(self.filepath, header = [3], encoding = 'ANSI', sep = ',')
        df['mm'] = df['mm'].apply(replace_negative_values)
        self.data = df
         
    def analyze(self, options):
        from .analyzers import StressStrain      
        from .fit import Fit
        from .plot import Plot
        
        data = StressStrain(self.data, sample_area=options['sample_area'], initial_length=options['initial_length'])
        print(data)
        fit = Fit(data)
        if "plot" not in options or options["plot"] is not False:
            Plot(fit, data, self.filename)
        return fit.results
    
def replace_negative_values(x):
    if x >= 0:
        return x
    if abs(x) < 0.01:
        return 0
    else:
        raise ValueError("Negative values in column 'mm'")    


class Plot:
    from .fit import Fit
    from .analyzers import StressStrain
    
    def __init__(self, fit: Fit, data: StressStrain, filename: str):
        self.fit_results = fit.results
        self.data = fit.data
        self.filename = filename
        self.plot()
        

    
    
    def plot(self):
        import matplotlib.pyplot as plt
        plt.plot(self.data['strain'], self.data['stress'])
        plt.xlabel('Strain')
        plt.ylabel('Stress [kPa]')
        plt.title(self.filename)
        
        self.plotBestFit()
        self.plotMaxStress()
        self.plotYield()
        
        plt.legend(['Data', 'Young Modulus', 'Shift', 'Max Stress', 'Yield'])
        
        plt.savefig('plots/'+self.filename + '.png')
        plt.draw()
        plt.pause(0.2)
        plt.close()
        
    def plotBestFit(self):
        import matplotlib.pyplot as plt
        max_stress = self.data['stress'].max()
        max_strain = (max_stress - self.fit_results['Intercept [kPa]']) / self.fit_results['Young Modulus [kPa]']
        max_strain = max_strain[0]
        strain_range = self.data[self.data['strain'] < max_strain]['strain']
        line = self.fit_results['Young Modulus [kPa]'][0] * strain_range + self.fit_results['Intercept [kPa]'][0]
        other = self.fit_results['Young Modulus [kPa]'][0] * (strain_range) + self.fit_results['Intercept [kPa]'][0]

        plt.plot(strain_range, line, linestyle='--')
        plt.plot(strain_range + 0.02, other, linestyle='dotted')
        
        
    def plotMaxStress(self):
        import matplotlib.pyplot as plt
        x = self.fit_results.iloc[0, 1]
        y = self.fit_results.iloc[0, 0]
        plt.scatter(x, y, marker='o', color='blue')
    
    def plotYield(self):
        import matplotlib.pyplot as plt
        x = self.fit_results.iloc[0, 5]
        y = self.fit_results.iloc[0, 4]
        plt.scatter(x, y, marker='o', color='green')
